
Require Import Tactics.
Require Import Sequence.
Require Import Subst.
Require Import Static.
Require Import AlgorithmEq.
Require Import Regularity.
Require Import Singleton.
Require Import Principal.


Definition aeqc G c1 c2 k := aeq G G c1 c2 k k.


Inductive asubk : context -> (* kd *) term -> term -> Prop :=
| asubk_type {G}
    : asubk G kd_type kd_type

| asubk_sing {G c c'}
    : aeqc G c c' kd_type
      -> asubk G (kd_sing c) (kd_sing c')

| asubk_sing_t {G c}
    : asubk G (kd_sing c) kd_type

| asubk_pi {G k1 k1' k2 k2'}
    : asubk G k1' k1
      -> asubk (G; cl_cn k1') k2 k2'
      -> asubk G (kd_pi k1 k2) (kd_pi k1' k2')

| asubk_sigma {G k1 k1' k2 k2'}
    : asubk G k1 k1'
      -> asubk (G; cl_cn k1) k2 k2'
      -> asubk G (kd_sigma k1 k2) (kd_sigma k1' k2')

| asubk_one {G}
    : asubk G kd_one kd_one
.


Inductive aofk : context -> (* kd *) term -> Prop :=
| aofk_type {G}
    : aofk G kd_type

| aofk_sing {G c}
    : aofc G c kd_type
      -> aofk G (kd_sing c)

| aofk_pi {G k1 k2}
    : aofk G k1
      -> aofk (G; cl_cn k1) k2
      -> aofk G (kd_pi k1 k2)

| aofk_sigma {G k1 k2}
    : aofk G k1
      -> aofk (G; cl_cn k1) k2
      -> aofk G (kd_sigma k1 k2)

| aofk_one {G}
    : aofk G kd_one


with aprink : context -> (* cn *) term -> (* kd *) term -> Prop :=
| aprink_var {G i k}
    : index i G (cl_cn k)
      -> aprink G (var i) (single (var i) (subst (sh (S i)) k))

| aprink_lam {G k1 k2 c}
    : aofk G k1
      -> aprink (G; cl_cn k1) c k2
      -> aprink G (cn_lam k1 c) (kd_pi k1 k2)

| aprink_app {G k1 k2 c1 c2}
    : aprink G c1 (kd_pi k1 k2)
      -> aofc G c2 k1
      -> aprink G (cn_app c1 c2) (subst1 c2 k2)

| aprink_pair {G k1 k2 c1 c2}
    : aprink G c1 k1
      -> aprink G c2 k2
      -> aprink G (cn_pair c1 c2) (kd_sigma k1 (shift1 k2))

| aprink_pi1 {G k1 k2 c}
    : aprink G c (kd_sigma k1 k2)
      -> aprink G (cn_pi1 c) k1

| aprink_pi2 {G k1 k2 c}
    : aprink G c (kd_sigma k1 k2)
      -> aprink G (cn_pi2 c) (subst1 (cn_pi1 c) k2)

| aprink_star {G}
    : aprink G cn_star kd_one

| aprink_unit {G}
    : aprink G cn_unit (kd_sing cn_unit)

| aprink_arrow {G c1 c2}
    : aofc G c1 kd_type
      -> aofc G c2 kd_type
      -> aprink G (cn_arrow c1 c2) (kd_sing (cn_arrow c1 c2))

| aprink_prod {G c1 c2}
    : aofc G c1 kd_type
      -> aofc G c2 kd_type
      -> aprink G (cn_prod c1 c2) (kd_sing (cn_prod c1 c2))

| aprink_all {G k c}
    : aofk G k
      -> aofc (G; cl_cn k) c kd_type
      -> aprink G (cn_all k c) (kd_sing (cn_all k c))

| aprink_exists {G k c}
    : aofk G k
      -> aofc (G; cl_cn k) c kd_type
      -> aprink G (cn_exists k c) (kd_sing (cn_exists k c))


with aofc : context -> (* cn *) term -> (* kd *) term -> Prop :=
| aofc_i {G c k k'}
    : aprink G c k'
      -> asubk G k' k
      -> aofc G c k
.


Scheme aofk_mut_ind := Minimality for aofk Sort Prop
with   aprink_mut_ind := Minimality for aprink Sort Prop
with   aofc_mut_ind := Minimality for aofc Sort Prop.
Combined Scheme aofk_aofc_ind from aofk_mut_ind, aprink_mut_ind, aofc_mut_ind.


Inductive aofs : context -> (* sg *) term -> Prop :=
| aofs_one {G}
    : aofs G sg_one

| aofs_stat {G k}
    : aofk G k
      -> aofs G (sg_stat k)

| aofs_dyn {G t}
    : aofc G t kd_type
      -> aofs G (sg_dyn t)

| aofs_pi {G T1 T2 k1}
    : aofs G T1
      -> fsts T1 k1
      -> aofs (G; cl_cn k1) T2
      -> aofs G (sg_pi T1 T2)

| aofs_sigma {G T1 T2 k1}
    : aofs G T1
      -> fsts T1 k1
      -> aofs (G; cl_cn k1) T2
      -> aofs G (sg_sigma T1 T2)

| aofs_circ {G P}
    : aofps G P
      -> aofs G (sg_circ P)

with aofps : context -> (* ps *) term -> Prop :=
| aofps_down {G T}
    : aofs G T
      -> aofps G (ps_down T)

| aofps_ex {G k P}
    : aofk G k
      -> aofps (G; cl_cn k) P
      -> aofps G (ps_ex k P)
.


Scheme aofs_mut_ind := Minimality for aofs Sort Prop
with   aofps_mut_ind := Minimality for aofps Sort Prop.
Combined Scheme aofsig_ind from aofs_mut_ind, aofps_mut_ind.


Inductive asubs : context -> (* sg *) term -> (* sg *) term -> Prop :=
| asubs_one {G}
    : asubs G sg_one sg_one

| asubs_stat {G k k'}
    : asubk G k k'
      -> asubs G (sg_stat k) (sg_stat k')

| asubs_dyn {G t t'}
    : aeqc G t t' kd_type
      -> asubs G (sg_dyn t) (sg_dyn t')

| asubs_pi {G T1 T1' k1' T2 T2'}
    : asubs G T1' T1
      -> fsts T1' k1'
      -> asubs (G; cl_cn k1') T2 T2'
      -> asubs G (sg_pi T1 T2) (sg_pi T1' T2')

| asubs_sigma {G T1 T1' k1 T2 T2'}
    : asubs G T1 T1'
      -> fsts T1 k1
      -> asubs (G; cl_cn k1) T2 T2'
      -> asubs G (sg_sigma T1 T2) (sg_sigma T1' T2')

| asubs_circ {G P P'}
    : asubps G P P'
      -> asubs G (sg_circ P) (sg_circ P')


with asubps : context -> (* ps *) term -> (* ps *) term -> Prop :=
| asubps_down {G T T'}
    : asubs G T T'
      -> asubps G (ps_down T) (ps_down T')

| asubps_ex_left {G k P P'}
    : asubps (G; cl_cn k) P (shift1 P')
      -> asubps G (ps_ex k P) P'
.


Scheme asubs_mut_ind := Minimality for asubs Sort Prop
with   asubps_mut_ind := Minimality for asubps Sort Prop.
Combined Scheme asubsig_ind from asubs_mut_ind, asubps_mut_ind.


Inductive close : context -> (* ps *) term -> (* ps *) term -> Prop :=
| close_nil {P}
    : close nil P P

| close_cons {G k P P'}
    : close G (ps_ex k P) P'
      -> close (G; cl_cn k) P P'.


Inductive aprint : context -> (* tm *) term -> (* cn *) term -> Prop :=
| aprint_var {G i t}
    : index i G (cl_tm t)
      -> aprint G (var i) (subst (sh (S i)) t)

| aprint_star {G}
    : aprint G tm_star cn_unit

| aprint_lam {G t1 t2 e}
    : aofc G t1 kd_type
      -> aprint (G; cl_tm t1) e t2
      -> aprint G (tm_lam t1 e) (cn_arrow t1 (subst1 nonsense t2))

| aprint_app {G t t1 t2 t1' e1 e2}
    : aprint G e1 t
      -> headnorm G t (cn_arrow t1 t2)
      -> aprint G e2 t1'
      -> aeqc G t1' t1 kd_type
      -> aprint G (tm_app e1 e2) t2

| aprint_pair {G t1 t2 e1 e2}
    : aprint G e1 t1
      -> aprint G e2 t2
      -> aprint G (tm_pair e1 e2) (cn_prod t1 t2)

| aprint_pi1 {G t t1 t2 e}
    : aprint G e t
      -> headnorm G t (cn_prod t1 t2)
      -> aprint G (tm_pi1 e) t1

| aprint_pi2 {G t t1 t2 e}
    : aprint G e t
      -> headnorm G t (cn_prod t1 t2)
      -> aprint G (tm_pi2 e) t2

| aprint_plam {G k t e}
    : aofk G k
      -> aprint (G; cl_cn k) e t
      -> svalue e
      -> aprint G (tm_plam k e) (cn_all k t)

| aprint_papp {G u k t e c}
    : aprint G e u
      -> headnorm G u (cn_all k t)
      -> aofc G c k
      -> aprint G (tm_papp e c) (subst1 c t)

| aprint_pack {G k t c e u}
    : aofk G k
      -> aofc G c k
      -> aprint G e u
      -> aeqc G u (subst1 c t) kd_type
      -> aofc (G; cl_cn k) t kd_type
      -> aprint G (tm_pack c e k t) (cn_exists k t)

| aprint_unpack {G u k t t' e1 e2}
    : aprint G e1 u
      -> headnorm G u (cn_exists k t)
      -> aprint (G; cl_cn k; cl_tm t) e2 t'
      -> aeqc (G; cl_cn k) (subst (dot nonsense id) t') (subst (dot nonsense (dot (shift1 (inh k)) sh1)) t') kd_type
      -> aprint G (tm_unpack e1 e2) (subst (dot nonsense (dot (inh k) id)) t')

| aprint_fix {G t e u}
    : aofc G t kd_type
      -> aprint G e u
      -> aeqc G u (cn_arrow (cn_arrow cn_unit t) t) kd_type
      -> aprint G (tm_fix t e) t

| aprint_lett {G t1 t2 e1 e2}
    : aprint G e1 t1
      -> aprint (G; cl_tm t1) e2 t2
      -> aprint G (tm_lett e1 e2) (subst1 nonsense t2)

| aprint_letm {G P G' T k t M e}
    : aprinps G M P
      -> linv P G' T
      -> fsts T k
      -> aprint (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) (subst (under 2 (sh (length G'))) e) t
      -> aeqc (G;; G'; cl_cn k) (subst (dot nonsense id) t) (subst (dot nonsense (dot (subst (compose (cinh G') (sh (S (length G')))) (inh k)) (compose (cinh G') (sh (S (length G')))))) t) kd_type
      -> aprint G (tm_letm M e) (subst (dot nonsense (dot (subst (cinh G') (inh k)) (cinh G'))) t)

| aprint_ext {G t M}
    : aprins G M (sg_dyn t)
      -> aprint G (tm_ext M) t


with aprins : context -> (* md *) term -> (* sg *) term -> Prop :=
| aprins_var {G i T c}
    : index i G (cl_md T c)
      -> aprins G (var i) (singles (subst (sh (S i)) c) (subst (sh (S i)) T))

| aprins_star {G}
    : aprins G md_star sg_one

| aprins_stat {G c k}
    : aprink G c k
      -> aprins G (md_stat c) (sg_stat k)

| aprins_dyn {G e t}
    : aprint G e t
      -> aprins G (md_dyn e) (sg_dyn t)

| aprins_lam {G T1 T2 k1 M}
    : aofs G T1
      -> fsts T1 k1
      -> aprins (G; cl_cn k1; cl_md (shift1 T1) (var 0)) M (shift1 T2)
      -> aprins G (md_lam T1 M) (sg_pi T1 T2)

| aprins_app {G T1 T2 T1' M1 M2 c2}
    : aprins G M1 (sg_pi T1 T2)
      -> aprins G M2 T1'
      -> asubs G T1' T1
      -> fstm G M2 c2
      -> aprins G (md_app M1 M2) (subst1 c2 T2)

| aprins_pair {G T1 T2 M1 M2}
    : aprins G M1 T1
      -> aprins G M2 T2
      -> aprins G (md_pair M1 M2) (sg_sigma T1 (shift1 T2))

| aprins_pi1 {G T1 T2 M}
    : aprins G M (sg_sigma T1 T2)
      -> aprins G (md_pi1 M) T1

| aprins_pi2 {G T1 T2 M c}
    : aprins G M (sg_sigma T1 T2)
      -> fstm G M c
      -> aprins G (md_pi2 M) (subst1 (cn_pi1 c) T2)

| aprins_lett {G t T e M}
    : aprint G e t
      -> aprins (G; cl_tm t) M T
      -> aprins G (md_lett e M) (subst1 nonsense T)

| aprins_cir {G M P}
    : aprinps G M P
      -> aprins G (md_cir M) (sg_circ P)


with aprinps : context -> (* lm *) term -> (* ps *) term -> Prop :=
| aprinps_ret {G M T}
    : aprins G M T
      -> aprinps G (lm_ret M) (ps_down T)

| aprinps_seal {G M T T'}
    : aprins G M T'
      -> aofs G T
      -> asubs G T' T
      -> aprinps G (lm_seal M T) (ps_down T)

| aprinps_bind {G M1 M2 P1 G' T k P2 P2'}
    : aprins G M1 (sg_circ P1)
      -> linv P1 G' T
      -> fsts T k
      -> aprinps (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) (subst (under 2 (sh (length G'))) M2) P2
      -> close G' (ps_ex k (subst1 nonsense P2)) P2'
      -> aprinps G (lm_bind M1 M2) P2'

| aprinps_unpack {G u k t P e M}
    : aprint G e u
      -> headnorm G u (cn_exists k t)
      -> aprinps (G; cl_cn k; cl_tm t) M P
      -> aprinps G (lm_unpack e M) (ps_ex k (subst1 nonsense P))
.


Scheme aprint_mut_ind := Minimality for aprint Sort Prop
with   aprins_mut_ind := Minimality for aprins Sort Prop
with   aprinps_mut_ind := Minimality for aprinps Sort Prop.
Combined Scheme aprin_ind from aprint_mut_ind, aprins_mut_ind, aprinps_mut_ind.


Inductive syn : (* sg *) term -> Prop :=
| syn_one 
    : syn sg_one

| syn_stat {k}
    : syn (sg_stat k)

| syn_dyn {t}
    : syn (sg_dyn t)

| syn_pi {T1 T2}
    : ana T1
      -> syn T2
      -> syn (sg_pi T1 T2)

| syn_sigma {T1 T2}
    : syn T1
      -> syn T2
      -> syn (sg_sigma T1 T2)

| syn_circ {P}
    : synp P
      -> syn (sg_circ P)

with synp : (* ps *) term -> Prop :=
| synp_down {T}
    : syn T
      -> synp (ps_down T)

| synp_ex {k P}
    : synp P
      -> synp (ps_ex k P)

with ana : (* sg *) term -> Prop :=
| ana_one 
    : ana sg_one

| ana_stat {k}
    : ana (sg_stat k)

| ana_dyn {t}
    : ana (sg_dyn t)

| ana_pi {T1 T2}
    : syn T1
      -> ana T2
      -> ana (sg_pi T1 T2)

| ana_sigma {T1 T2}
    : ana T1
      -> ana T2
      -> ana (sg_sigma T1 T2)

| ana_circ {P}
    : anap P
      -> ana (sg_circ P)

with anap : (* ps *) term -> Prop :=
| anap_down {T}
    : ana T
      -> anap (ps_down T)
.


Scheme syn_mut_ind := Minimality for syn Sort Prop
with   synp_mut_ind := Minimality for synp Sort Prop
with   ana_mut_ind := Minimality for ana Sort Prop
with   anap_mut_ind := Minimality for anap Sort Prop.
Combined Scheme syn_ana_ind from syn_mut_ind, synp_mut_ind, ana_mut_ind, anap_mut_ind.


Inductive synctx : context -> Prop :=
| synctx_nil
    : synctx nil

| synctx_cn {G k}
    : synctx G
      -> synctx (G; cl_cn k)

| synctx_tm {G t}
    : synctx G
      -> synctx (G; cl_tm t)

| synctx_md {G T c}
    : synctx G
      -> syn T
      -> synctx (G; cl_md T c)
.


Inductive noex : term -> Prop :=
| noex_one
    : noex sg_one

| noex_stat {k}
    : noex (sg_stat k)

| noex_dyn {t}
    : noex (sg_dyn t)

| noex_pi {T1 T2}
    : noex T1
      -> noex T2
      -> noex (sg_pi T1 T2)

| noex_sigma {T1 T2}
    : noex T1
      -> noex T2
      -> noex (sg_sigma T1 T2)

| noex_circ {P}
    : noexp P
      -> noex (sg_circ P)


with noexp : term -> Prop :=
| noexp_down {T}
    : noex T
      -> noexp (ps_down T)
.


Scheme noex_mut_ind := Minimality for noex Sort Prop
with   noexp_mut_ind := Minimality for noexp Sort Prop.
Combined Scheme noexx_ind from noex_mut_ind, noexp_mut_ind.


Inductive user : term -> Prop :=
| user_var {i}
    : user (var i)

| user_tm_star
    : user tm_star

| user_tm_lam {t e}
    : user e
      -> user (tm_lam t e)

| user_tm_app {e1 e2}
    : user e1
      -> user e2
      -> user (tm_app e1 e2)

| user_tm_pair {e1 e2}
    : user e1
      -> user e2
      -> user (tm_pair e1 e2)

| user_tm_pi1 {e}
    : user e
      -> user (tm_pi1 e)

| user_tm_pi2 {e}
    : user e
      -> user (tm_pi2 e)

| user_tm_plam {k e}
    : user e
      -> user (tm_plam k e)

| user_tm_papp {e c}
    : user e
      -> user (tm_papp e c)

| user_tm_pack {k t c e}
    : user e
      -> user (tm_pack c e k t)

| user_tm_unpack {e1 e2}
    : user e1
      -> user e2
      -> user (tm_unpack e1 e2)

| user_tm_fix {t e}
    : user e
      -> user (tm_fix t e)

| user_tm_lett {e1 e2}
    : user e1
      -> user e2
      -> user (tm_lett e1 e2)

| user_tm_letm {M e}
    : user M
      -> user e
      -> user (tm_letm M e)

| user_tm_ext {M}
    : user M
      -> user (tm_ext M)

| user_md_star
    : user md_star

| user_md_stat {c}
    : user (md_stat c)

| user_md_dyn {e}
    : user e
      -> user (md_dyn e)

| user_md_lam {T M}
    : noex T
      -> user M
      -> user (md_lam T M)

| user_md_app {M1 M2}
    : user M1
      -> user M2
      -> user (md_app M1 M2)

| user_md_pair {M1 M2}
    : user M1
      -> user M2
      -> user (md_pair M1 M2)

| user_md_pi1 {M}
    : user M
      -> user (md_pi1 M)

| user_md_pi2 {M}
    : user M
      -> user (md_pi2 M)

| user_md_lett {e M}
    : user e
      -> user M
      -> user (md_lett e M)

| user_md_cir {M}
    : user M
      -> user (md_cir M)

| user_lm_ret {M}
    : user M
      -> user (lm_ret M)

| user_lm_seal {M T}
    : user M
      -> noex T
      -> user (lm_seal M T)

| user_lm_bind {M1 M2}
    : user M1
      -> user M2
      -> user (lm_bind M1 M2)

| user_lm_unpack {e M}
    : user e
      -> user M
      -> user (lm_unpack e M)
.
