  
Require Import Decidable.
Require Import Wf_nat.
Require Import Compare_dec.

Require Import Tactics.
Require Import Sequence.
Require Import Subst.
Require Import Static.
Require Import TSubst.
Require Import Regularity.
Require Import Formation.
Require Import Singleton.
Require Import Declarative.
Require Import Preservation.
Require Import AlgorithmEq.
Require Import SubjectReduction.
Require Import Algorithm.
Require Import SoundEq.
Require Import Sound.
Require Import CompleteEq.
Require Import Complete.
Require Import Principal.
Require Import Simp.


Lemma dec_equiv :
  forall (P Q : Prop),
    (P -> Q)
    -> (Q -> P)
    -> decidable P
    -> decidable Q.
Proof.
intros P Q HPQ HQP Hdec.
destruct Hdec as [| H].
  {
  left; auto.
  }

  {
  right.
  contradict H.
  auto.
  }
Qed.


Lemma dec_chain :
  forall A (P : A -> Prop) (Q : A -> Prop),
    (forall x y, P x -> P y -> x = y)
    -> decidable (exists x, P x)
    -> (forall x, P x -> decidable (Q x))
    -> decidable (exists x, P x /\ Q x).
Proof.
intros A P Q Hfun Hdec1 Hdec2.
destruct Hdec1 as [(x & H1) | Hno].
  {
  destruct (Hdec2 x H1) as [H2 | Hno].
    {
    left; eauto.
    }
  right.
  contradict Hno.
  destruct Hno as (y & H1' & H2).
  so (Hfun _ _ H1 H1').
  subst y.
  exact H2.
  }
right.
contradict Hno.
destruct Hno as (x & H1 & _).
eauto.
Qed.


Lemma dec_chain2 :
  forall A B (P : A -> B -> Prop) (Q : A -> B -> Prop),
    (forall x x' y y', P x y -> P x' y' -> x = x' /\ y = y')
    -> decidable (exists x y, P x y)
    -> (forall x y, P x y -> decidable (Q x y))
    -> decidable (exists x y, P x y /\ Q x y).
Proof.
intros A B P Q Hfun Hdec1 Hdec2.
destruct Hdec1 as [(x & y & H1) | Hno].
  {
  destruct (Hdec2 x y H1) as [H2 | Hno].
    {
    left; eauto.
    }
  right.
  contradict Hno.
  destruct Hno as (x' & y' & H1' & H2).
  destruct (Hfun _ _ _ _ H1 H1').
  subst x' y'.
  exact H2.
  }
right.
contradict Hno.
destruct Hno as (x & x' & H1 & _).
eauto.
Qed.


Lemma dec_and_dep :
  forall A B,
    decidable A
    -> (A -> decidable B)
    -> decidable (A /\ B).
Proof.
intros A B HA HB.
destruct HA as [HA | Hno].
  {
  destruct (HB HA) as [? | Hno].
    {
    left; auto.
    }
  right; contradict Hno; destruct Hno; auto.
  }
right; contradict Hno; destruct Hno; auto.
Qed.


Ltac destruct_all_or_main H :=
  (lazymatch type of H with
   | exists _:_, _ => 
               let H' := fresh "x"
               in
                 destruct H as [H' H]; destruct_all_or_main H'; destruct_all_or_main H
   | _ /\ _ => let H' := fresh "H"
               in
                 destruct H as [H' H]; destruct_all_or_main H'; destruct_all_or_main H
   | prod _ _ =>
               let H' := fresh "x"
               in
                 destruct H as [H' H]; destruct_all_or_main H'; destruct_all_or_main H
   | _ \/ _ => let H1 := fresh "H" in
               let H2 := fresh "H"
               in
                 destruct H as [H1 | H2]; [destruct_all_or_main H1 | destruct_all_or_main H2]
   | _ => idtac
   end).


Ltac destruct_all_or t :=
  first [
        destruct_all_or_main t
        |
        let H := fresh
        in
          pose proof t as H;
          destruct_all_or_main H
        ].


Ltac decform m :=
  let H := fresh
  in
    destruct m; try (right; intro H; destruct_all H; discriminate); left; eauto; done.
  

Ltac into t :=
  let H := fresh
  in
    intro H; destruct_all_or H; subst; eauto using t; done.


Ltac from :=
  let H := fresh
  in
    intro H; destruct_all H; invert H; intros; subst; eauto 8; done.


Ltac rephrase t P :=
  apply (dec_equiv P); [into t | from |].


Ltac injfun :=
  intros;
  subst;
  match goal with
  | H : _ = _ |- _ => injection H; intros; subst; auto; done
  end.


Lemma index_decide :
  forall T i (l : list T),
    decidable (exists x, index i l x).
Proof.
intros T i l.
revert l.
induct i.

(* 0 *)
{
intros l.
destruct l as [| x l].
  {
  right.
  intros (x & H).
  invert H.
  }
left.
exists x.
apply index_0.
}

(* S *)
{
intros i IH l.
destruct l as [| x l].
  {
  right.
  intros (x & H).
  invert H.
  }
so (IH l) as [(y & Hindex) | H].
  {
  left.
  exists y.
  apply index_S; auto.
  }
right.
contradict H.
destruct H as (y & Hindex).
invert Hindex.
eauto.
}
Qed.


Local Ltac inject_decidable m n :=
  match m with
  | ?m1 ?m2 =>
      match n with
      | ?n1 ?n2 =>
          let H := fresh
          in
            assert (decidable (m2 = n2)) as H;
            [ auto 
            | destruct H as [H | H];
              [ inject_decidable m1 n1
              | right; contradict H; injection H; auto
              ]
            ]
        
      | _ => idtac
      end

  | _ => idtac
  end.


Lemma dec_eq_term :
  forall (m n : term), decidable (m = n).
Proof.
intros m n.
revert n.
induct m;
try (intros; 
     destruct n; 
     try (right; intro HH; discriminate HH);
     left; reflexivity);
try (intros;
     destruct n; try (right; intro HH; discriminate HH);
     match goal with
     | |- decidable (?m = ?n) => inject_decidable m n
     end; left; f_equal; auto; done).

(* var *)
{
intros i n.
destruct n; try (right; intro HH; discriminate HH).
apply (dec_equiv (i = n)); auto.
  {
  intro H; injection H; auto.
  }
apply Peano_dec.dec_eq_nat.
}
Qed.


Lemma aeq_decide :
  forall G c d k,
    wf G
    -> ofc G c k
    -> ofc G d k
    -> decidable (aeq G G c d k k).
Proof.
exploit
  (alg_ind
     (fun G G' c c' k k' =>
        forall C d l,
          aeq C C d d l l
          -> decidable (aeq G C c d k l))
     (fun G G' p p' k k' =>
        forall C q l,
          peq C C q q l l
          -> decidable (peq G C p q k l))
     (fun G G' k k' =>
        forall C l,
          aeqk C C l l
          -> decidable (aeqk G C k l))) as Hind.

(* type *)
{
intros G _ c _ p _ Hnormc _ _ IH C d l Haeqd.
assert (decidable (l = kd_type)) as [-> | Hno] by (decform l).
  {
  invert Haeqd; [].
  intros q q' Hnormd Hnormd' Hpeqq.
  so (headnorm_fun _#4 Hnormd Hnormd'); subst q'; clear Hnormd'.
  so (IH _#3 Hpeqq) as Hd.
  eapply dec_equiv; [| | exact Hd].
    {
    intros Hpeq.
    eapply aeq_type; eauto.
    }

    {
    intro Haeq.
    invert Haeq.
    intros p' q' Hnormc' Hnormd' Hpeq.
    so (headnorm_fun _#4 Hnormc Hnormc'); subst p'.
    so (headnorm_fun _#4 Hnormd Hnormd'); subst q'.
    exact Hpeq.
    }
  }

  {
  right.
  contradict Hno.
  invert Hno; auto.
  }
}

(* sing *)
{
intros G _ c _ b _ C d l _.
rephrase aeq (exists a, l = kd_sing a).
decform l.
}

(* pi *)
{
intros G _ c _ k1 _ k2 _ _ IH C d l Haeqd.
assert (decidable (exists l1 l2, l = kd_pi l1 l2)) as [(l1 & l2 & ->) | Hno] by (decform l).
  {
  invert Haeqd; [].
  intros Haeqdx.
  so (IH _ _ _ Haeqdx) as Hd.
  eapply dec_equiv; [| | exact Hd].
    {
    eauto using aeq.
    }
    
    {
    intro Haeq.
    invert Haeq; eauto.
    }
  }

  {
  right.
  contradict Hno.
  invert Hno; eauto.
  }
}

(* sigma *)
{
intros G _ c _ k1 _ k2 _ _ IH1 _ IH2 C d l Haeqd.
assert (decidable (exists l1 l2, l = kd_sigma l1 l2)) as [(l1 & l2 & ->) | Hno] by (decform l).
  {
  invert Haeqd; [].
  intros Haeqd1 Haeqd2.
  so (IH1 _ _ _ Haeqd1) as Hd1.
  so (IH2 _ _ _ Haeqd2) as Hd2.
  eapply dec_equiv; [| | exact (dec_and _ _ Hd1 Hd2)].
    {
    intros (H1 & H2).
    eauto using aeq.
    }
    
    {
    intro Haeq.
    invert Haeq; eauto.
    }
  }

  {
  right.
  contradict Hno.
  invert Hno; eauto.
  }
}

(* one *)
{
intros G _ c _ C d l _.
rephrase aeq (l = kd_one).
decform l.
}

(* var *)
{
intros G _ i k _ Hindex _ C q l Hpeqq.
apply (dec_equiv (q = var i /\ exists m, index i C m /\ exists m', m = cl_cn m' /\ l = subst (sh (S i)) m')).
  { into peq. }

  { from. }
apply dec_and; auto using dec_eq_term.
apply dec_chain; eauto using index_fun, index_decide.
intros m _.
apply dec_chain; [injfun | decform m |].
intros m' _.
apply dec_eq_term.
}

(* unit *)
{
intros G _ C q l _.
rephrase peq (q = cn_unit /\ l = kd_type).
apply dec_and; auto using dec_eq_term.
}

(* arrow *)
{
intros G _ t1 _ t2 _ _ IH1 _ IH2 C q l Hpeqq.
rephrase peq (exists u1 u2, q = cn_arrow u1 u2 /\ l = kd_type /\ aeq G C t1 u1 kd_type kd_type /\ aeq G C t2 u2 kd_type kd_type).
apply dec_chain2.
  { injfun. }

  { decform q. }
intros u1 u2 ->.
apply dec_and_dep; auto using dec_eq_term.
intros ->.
invert Hpeqq.
intros Haeq1 Haeq2.
apply dec_and; auto.
}

(* arrow *)
{
intros G _ t1 _ t2 _ _ IH1 _ IH2 C q l Hpeqq.
rephrase peq (exists u1 u2, q = cn_prod u1 u2 /\ l = kd_type /\ aeq G C t1 u1 kd_type kd_type /\ aeq G C t2 u2 kd_type kd_type).
apply dec_chain2.
  { injfun. }

  { decform q. }
intros u1 u2 ->.
apply dec_and_dep; auto using dec_eq_term.
intros ->.
invert Hpeqq.
intros Haeq1 Haeq2.
apply dec_and; auto.
}

(* all *)
{
intros G _ t1 _ t2 _ _ IH1 _ IH2 C q l Hpeqq.
rephrase peq (exists u1 u2, q = cn_all u1 u2 /\ l = kd_type /\ aeqk G C t1 u1 /\ aeq (G; cl_cn t1) (C; cl_cn u1) t2 u2 kd_type kd_type).
apply dec_chain2.
  { injfun. }

  { decform q. }
intros u1 u2 ->.
apply dec_and_dep; auto using dec_eq_term.
intros ->.
invert Hpeqq.
intros Haeq1 Haeq2.
apply dec_and; auto.
}

(* exists *)
{
intros G _ t1 _ t2 _ _ IH1 _ IH2 C q l Hpeqq.
rephrase peq (exists u1 u2, q = cn_exists u1 u2 /\ l = kd_type /\ aeqk G C t1 u1 /\ aeq (G; cl_cn t1) (C; cl_cn u1) t2 u2 kd_type kd_type).
apply dec_chain2.
  { injfun. }

  { decform q. }
intros u1 u2 ->.
apply dec_and_dep; auto using dec_eq_term.
intros ->.
invert Hpeqq.
intros Haeq1 Haeq2.
apply dec_and; auto.
}

(* app *)
{
intros G G' p p' c c' k1 k1' k2 k2' Hpeqp IH1 _ IH2 C q l Hpeqq.
assert (decidable (exists r d, q = cn_app r d)) as [(r & d & ->) | Hno] by (decform q).
  {
  invert Hpeqq.
  intros l1 l1' l2 l2' Hpeqr Haeqd <- _.
  so (peq_self_left _#6 Hpeqr) as Hpeqr'.
  so (aeq_self_left _#6 Haeqd) as Haeqd'.
  apply (dec_equiv (peq G C p r (kd_pi k1 k2) (kd_pi l1 l2) /\ aeq G C c d k1 l1)).
    { into peq. }

    {
    intros Hpeq.
    invert Hpeq.
    intros k1'' l1'' k2'' l2'' Hpeqpr Haeq _ _.
    so (natk_fun _#4 (peq_natk _#6 Hpeqp andel) (peq_natk _#6 Hpeqpr andel)) as H.
    injection H; clear H.
    intros <- <-.
    so (natk_fun _#4 (peq_natk _#6 Hpeqr andel) (peq_natk _#6 Hpeqpr ander)) as H.
    injection H; clear H.
    intros <- <-.
    auto.
    }
  apply dec_and; auto.
  }

  {
  right.
  contradict Hno.
  invert Hno; intros; subst; eauto.
  }
}

(* pi1 *)
{
intros G G' p p' k1 k1' k2 k2' Heqp IH C q l Heqq.
assert (decidable (exists r, q = cn_pi1 r)) as [(r & ->) | Hno] by (decform q).
  {
  invert Heqq.
  intros l2 l2' Heqr.
  apply (dec_equiv (peq G C p r (kd_sigma k1 k2) (kd_sigma l l2))).
    { into peq. }

    {
    intros Hpeq.
    invert Hpeq.
    intros k2'' l2'' Heqpr.
    so (natk_fun _#4 (peq_natk _#6 Heqp andel) (peq_natk _#6 Heqpr andel)) as H.
    injection H; clear H.
    intros <-.
    so (natk_fun _#4 (peq_natk _#6 Heqr andel) (peq_natk _#6 Heqpr ander)) as H.
    injection H; clear H.
    intros <-.
    auto.
    }
  apply IH.
  eapply peq_self_left; eauto.
  }

  {
  right.
  contradict Hno.
  invert Hno; intros; subst; eauto.
  }
}

(* pi2 *)
{
intros G G' p p' k1 k1' k2 k2' Heqp IH C q l Heqq.
assert (decidable (exists r, q = cn_pi2 r)) as [(r & ->) | Hno] by (decform q).
  {
  invert Heqq.
  intros l1 l1' l2 l2' Heqr <- _.
  apply (dec_equiv (peq G C p r (kd_sigma k1 k2) (kd_sigma l1 l2))).
    { into peq. }

    {
    intros Hpeq.
    invert Hpeq.
    intros k1'' k2'' l1'' l2'' Heqpr _ _.
    so (natk_fun _#4 (peq_natk _#6 Heqp andel) (peq_natk _#6 Heqpr andel)) as H.
    injection H; clear H.
    intros <- <-.
    so (natk_fun _#4 (peq_natk _#6 Heqr andel) (peq_natk _#6 Heqpr ander)) as H.
    injection H; clear H.
    intros <- <-.
    auto.
    }
  apply IH.
  eapply peq_self_left; eauto.
  }

  {
  right.
  contradict Hno.
  invert Hno; intros; subst; eauto.
  }
}

(* star *)
{
intros G _ C q l _.
rephrase peq (q = cn_star /\ l = kd_one).
apply dec_and; auto using dec_eq_term.
}

(* type *)
{
intros G _ C l _.
rephrase aeqk (l = kd_type).
apply dec_eq_term.
}

(* sing *)
{
intros G _ c _ _ IH C l Heql.
rephrase aeqk (exists d, l = kd_sing d /\ aeq G C c d kd_type kd_type).
apply dec_chain; [injfun | decform l |].
intros d ->.
apply IH.
invert Heql; auto.
}

(* pi *)
{
intros G _ k1 _ k2 _ _ IH1 _ IH2 C l Heql.
rephrase aeqk (exists l1 l2, l = kd_pi l1 l2 /\ aeqk G C k1 l1 /\ aeqk (G; cl_cn k1) (C; cl_cn l1) k2 l2).
apply dec_chain2; [injfun | decform l |].
intros l1 l2 ->.
invert Heql; intros.
apply dec_and; auto.
}

(* sigma *)
{
intros G _ k1 _ k2 _ _ IH1 _ IH2 C l Heql.
rephrase aeqk (exists l1 l2, l = kd_sigma l1 l2 /\ aeqk G C k1 l1 /\ aeqk (G; cl_cn k1) (C; cl_cn l1) k2 l2).
apply dec_chain2; [injfun | decform l |].
intros l1 l2 ->.
invert Heql; intros.
apply dec_and; auto.
}

(* one *)
{
intros G _ C l _.
rephrase aeqk (l = kd_one).
apply dec_eq_term.
}

(* epilogue *)
{
destruct Hind as (Hdec & _).
intros G c d k Hwf Hofc Hofd.
so (aeq_complete _#4 Hwf (eqc_refl Hofc)).
so (aeq_complete _#4 Hwf (eqc_refl Hofd)).
eapply Hdec; eauto.
}
Qed.


Lemma aeqc_decide :
  forall G c d k,
    wf G
    -> ofc G c k
    -> ofc G d k
    -> decidable (aeqc G c d k).
Proof.
intros G c d k Hwf Hofc Hofd.
unfold aeqc.
so (aeq_decide _#4 Hwf Hofc Hofd) as [Hyes | Hno].
  {
  left; auto.
  }

  {
  right.
  contradict Hno.
  auto.
  }
Qed.



Lemma eqc_decide :
  forall G c d k,
    wf G
    -> ofc G c k
    -> ofc G d k
    -> decidable (eqc G c d k).
Proof.
intros G c d k Hwf Hofc Hofd.
apply (dec_equiv (aeqc G c d k)); auto using aeqc_sound, aeqc_complete, aeqc_decide.
Qed.


Lemma dec_eq_skind :
  forall (sk sk' : skind), decidable (sk = sk').
Proof.
intros m n.
revert n.
induct m;
try (intros; 
     destruct n; 
     try (right; intro HH; discriminate HH);
     left; reflexivity);
try (intros;
     destruct n; try (right; intro HH; discriminate HH);
     match goal with
     | |- decidable (?m = ?n) => inject_decidable m n
     end; left; f_equal; auto; done).
Qed.


Lemma asubk_decide :
  forall G k k',
    wf G
    -> ofk G k
    -> ofk G k'
    -> decidable (asubk G k k').
Proof.
intros G k k' Hwf Hofk Hofk'.
cut (forall sk, simp k = sk -> simp k' = sk -> decidable (asubk G k k')).
  {
  intro Hind.
  so (dec_eq_skind (simp k) (simp k')) as [Heq | Hno].
    {
    eapply Hind; eauto.
    }
  right.
  contradict Hno.
  eapply subk_simp; eauto using asubk_sound.
  }
intros sk.
revert G k k' Hwf Hofk Hofk'.
induct sk.

(* type/sing *)
{
intros G k k' Hwf Hofk Hofk' Hsk Hsk'.
invert Hofk; try (intros; subst; discriminate Hsk).
  {
  intros <-.
  invert Hofk'; try (intros; subst; discriminate Hsk').
    {
    intros <-.
    left.
    apply asubk_type.
    }

    {
    intros c _ <-.
    right.
    intro H.
    invert H.
    }
  }

  {
  intros c Hofc <-.
  invert Hofk'; try (intros; subst; discriminate Hsk').
    {
    intros <-.
    left.
    apply asubk_sing_t.
    }

    {
    intros c' Hofc' <-.
    so (aeq_decide _#4 Hwf Hofc Hofc') as [Hyes | Hno].
      {
      left.
      apply asubk_sing; auto.
      }

      {
      right.
      contradict Hno.
      invert Hno; auto.
      }
    }
  }
}

(* pi *)
{
intros sk1 IH1 sk2 IH2 G k k' Hwf Hofk Hofk' Hsk Hsk'.
invert Hofk; try (intros; subst; discriminate Hsk).
intros k1 k2 Hofk1 Hofk2 <-.
invert Hofk'; try (intros; subst; discriminate Hsk').
intros k1' k2' Hofk1' Hofk2' <-.
injection Hsk.
intros Hsk2 Hsk1.
injection Hsk'.
intros Hsk2' Hsk1'.
rephrase asubk (asubk G k1' k1 /\ asubk (G; cl_cn k1') k2 k2').
apply dec_and_dep; auto.
intros Hasub.
apply IH2; auto.
  {
  apply wf_cn; auto.
  }

  {
  so (asubk_sound _#3 Hwf Hofk1' Hofk1 Hasub) as Hsub.
  so (subst_ofk _#4 (subof_subsume_0 _#3 Hsub) Hofk2) as H.
  simp_sub in H.
  exact H.
  }
}

(* sigma *)
{
intros sk1 IH1 sk2 IH2 G k k' Hwf Hofk Hofk' Hsk Hsk'.
invert Hofk; try (intros; subst; discriminate Hsk).
intros k1 k2 Hofk1 Hofk2 <-.
invert Hofk'; try (intros; subst; discriminate Hsk').
intros k1' k2' Hofk1' Hofk2' <-.
injection Hsk.
intros Hsk2 Hsk1.
injection Hsk'.
intros Hsk2' Hsk1'.
rephrase asubk (asubk G k1 k1' /\ asubk (G; cl_cn k1) k2 k2').
apply dec_and_dep; auto.
intros Hasub.
apply IH2; auto.
  {
  apply wf_cn; auto.
  }

  {
  so (asubk_sound _#3 Hwf Hofk1 Hofk1' Hasub) as Hsub.
  so (subst_ofk _#4 (subof_subsume_0 _#3 Hsub) Hofk2') as H.
  simp_sub in H.
  exact H.
  }
}

(* one *)
{
intros G k k' Hwf Hofk Hofk' Hsk Hsk'.
invert Hofk; try (intros; subst; discriminate Hsk).
intros <-.
invert Hofk'; try (intros; subst; discriminate Hsk').
intros <-.
left.
apply asubk_one.
}
Qed.


Lemma subk_decide :
  forall G k k',
    wf G
    -> ofk G k
    -> ofk G k'
    -> decidable (subk G k k').
Proof.
intros G k k' Hwf Hofk Hofk'.
apply (dec_equiv (asubk G k k')); auto using asubk_decide, asubk_sound, asubk_complete.
Qed.


Lemma aprink_fun :
  forall G c k k',
    wf G
    -> aprink G c k
    -> aprink G c k'
    -> k = k'.
Proof.
intros G c k k' Hwf Hprink Hprink'.
eapply prink_fun; eauto using aprink_sound_pre.
Qed.


Lemma aprink_dec_impl :
  forall G c k, 
    wf G
    -> ofk G k
    -> decidable (exists k', aprink G c k') 
    -> decidable (aofc G c k).
Proof.
intros G c k Hwf Hofk Hdec.
destruct Hdec as [(k' & Hprink) | Hno].
  {
  so (ofc_reg _#3 Hwf (aprink_sound _#3 Hwf Hprink)) as Hofk'.
  apply (dec_equiv (asubk G k' k)); eauto using asubk_decide, aofc_i.
  intro Haofc.
  invert Haofc.
  intros k'' Hprink' Hsub.
  so (aprink_fun _#4 Hwf Hprink Hprink').
  subst k''.
  exact Hsub.
  }

  {
  right.
  contradict Hno.
  invert Hno; eauto.
  }
Qed.


Lemma aofk_aprink_decide :
  (forall G k,
     wf G
     -> decidable (aofk G k))
  /\
  (forall G c,
     wf G
     -> decidable (exists k, aprink G c k)).
Proof.
cut (forall m,
       (forall G, wf G -> decidable (aofk G m))
       /\
       (forall G, wf G -> decidable (exists k, aprink G m k))).
  {
  intros Hind.
  split.
    {
    intros G k.
    apply (Hind k andel).
    }

    {
    intros G c.
    apply (Hind c ander).
    }
  }
intro m.
induct m;
intros; split;
try (intros; right; intro HH; destruct_all HH; invert HH; done);
repeat
  (match goal with
   | H : _ /\ _ |- _ => destruct H
   end);
revert_all;
try (left; eauto using aofk, aprink; done).

(* var *)
{
intros i G Hwf.
rephrase aprink (exists A, index i G A /\ exists k, A = cl_cn k).
apply dec_chain; eauto using index_fun, index_decide.
intros A Hindex.
decform A.
}

(* sing *)
{
intros c _ IH G Hwf.
rephrase aofk (aofc G c kd_type).
apply aprink_dec_impl; auto with static.
}

(* pi *)
{
intros k1 IH1 _ k2 IH2 _ G Hwf.
rephrase aofk (aofk G k1 /\ aofk (G; cl_cn k1) k2).
apply dec_and_dep; auto.
intro Hofk1.
apply IH2; auto using wf, aofk_sound.
}

(* sigma *)
{
intros k1 IH1 _ k2 IH2 _ G Hwf.
rephrase aofk (aofk G k1 /\ aofk (G; cl_cn k1) k2).
apply dec_and_dep; auto.
intro Hofk1.
apply IH2; auto using wf, aofk_sound.
}

(* lam *)
{
intros k1 IH1 _ c _ IH2 G Hwf.
rephrase aprink (aofk G k1 /\ exists k2, aprink (G; cl_cn k1) c k2).
apply dec_and_dep; auto.
intros Hofk1.
apply IH2; auto using wf, aofk_sound.
}

(* app *)
{
intros c1 _ IH1 c2 _ IH2 G Hwf.
rephrase aprink (exists k, aprink G c1 k /\ exists k1 k2, k = kd_pi k1 k2 /\ aofc G c2 k1).
apply dec_chain; eauto using aprink_fun.
intros k Hprink.
apply dec_chain2; [injfun | decform k |].
intros k1 k2 ->.
apply aprink_dec_impl; auto.
invert (ofc_reg _#3 Hwf (aprink_sound _#3 Hwf Hprink)); auto.
}

(* pair *)
{
intros c1 _ IH1 c2 _ IH2 G Hwf.
rephrase aprink ((exists k1, aprink G c1 k1) /\ (exists k2, aprink G c2 k2)).
apply dec_and; auto.
}

(* pi1 *)
{
intros c _ IH G Hwf.
rephrase aprink (exists k, aprink G c k /\ exists k1 k2, k = kd_sigma k1 k2).
apply dec_chain; eauto using aprink_fun.
intros k _.
decform k.
}

(* pi2 *)
{
intros c _ IH G Hwf.
rephrase aprink (exists k, aprink G c k /\ exists k1 k2, k = kd_sigma k1 k2).
apply dec_chain; eauto using aprink_fun.
intros k _.
decform k.
}

(* arrow *)
{
intros t1 _ IH1 t2 _ IH2 G Hwf.
rephrase aprink (aofc G t1 kd_type /\ aofc G t2 kd_type).
apply dec_and; apply aprink_dec_impl; auto using ofk_type.
}

(* prod *)
{
intros t1 _ IH1 t2 _ IH2 G Hwf.
rephrase aprink (aofc G t1 kd_type /\ aofc G t2 kd_type).
apply dec_and; apply aprink_dec_impl; auto using ofk_type.
}

(* all *)
{
intros k IH1 _ t _ IH2 G Hwf.
rephrase aprink (aofk G k /\ aofc (G; cl_cn k) t kd_type).
apply dec_and_dep; auto.
intro Hofk.
apply aprink_dec_impl; eauto using aofk_sound with static.
}

(* exists *)
{
intros k IH1 _ t _ IH2 G Hwf.
rephrase aprink (aofk G k /\ aofc (G; cl_cn k) t kd_type).
apply dec_and_dep; auto.
intro Hofk.
apply aprink_dec_impl; eauto using aofk_sound with static.
}
Qed.


Lemma aofk_decide :
  forall G k,
    wf G
    -> decidable (aofk G k).
Proof.
exact (aofk_aprink_decide andel).
Qed.


Lemma aprink_decide :
  forall G c,
    wf G
    -> decidable (exists k, aprink G c k).
Proof.
exact (aofk_aprink_decide ander).
Qed.


Lemma ofk_decide :
  forall G k,
    wf G
    -> decidable (ofk G k).
Proof.
intros G k Hwf.
apply (dec_equiv (aofk G k)); eauto using aofk_decide, aofk_sound, aofk_complete.
Qed.


Lemma aofc_decide :
  forall G c k,
    wf G
    -> ofk G k
    -> decidable (aofc G c k).
Proof.
intros G c k Hwf Hofk.
apply (dec_equiv (exists k', aprink G c k' /\ asubk G k' k)).
  { into aofc. }

  { from. }
apply dec_chain; eauto using aprink_fun, aprink_decide.
intros k' Hprink.
so (ofc_reg _#3 Hwf (aprink_sound _#3 Hwf Hprink)) as Hofk'.
apply asubk_decide; auto.
Qed.


Lemma ofc_decide :
  forall G c k,
    wf G
    -> decidable (ofc G c k).
Proof.
intros G c k Hwf.
apply (dec_equiv (aofk G k /\ aofc G c k)).
  {
  intros (Haofk & Haofc).
  eauto using aofk_sound, aofc_sound.
  }
  
  {
  intro Hofc.
  so (ofc_reg _#3 Hwf Hofc) as Hofk.
  auto using aofk_complete, aofc_complete.
  }
apply dec_and_dep; auto using aofk_decide.
intro Hofk.
apply aofc_decide; auto using aofk_sound.
Qed.


Lemma aofx_decide :
  forall G T,
    wf G
    -> decidable (aofs G T) /\ decidable (aofps G T).
Proof.
intros G T Hwf.
revert G Hwf.
induct T;
intros;
split;
try (right; intro HH; invert HH; done);
revert_all.

(* one *)
{
left; eauto using aofs.
}

(* stat *)
{
intros k _ G Hwf.
rephrase aofs (aofk G k).
apply aofk_decide; auto.
}

(* dyn *)
{
intros t _ G Hwf.
rephrase aofs (aofc G t kd_type).
apply aofc_decide; auto with static.
}

(* pi *)
{
intros T1 IH1 T2 IH2 G Hwf.
rephrase aofs (aofs G T1 /\ exists k, fsts T1 k /\ aofs (G; cl_cn k) T2).
apply dec_and_dep; [apply IH1; auto |].
intros HaofT1.
so (aofs_sound _ _ Hwf HaofT1) as HofT1.
apply dec_chain; eauto using fsts_fun.
  {
  so (can_fsts _ _ HofT1) as (k & Hfst).
  left; eauto.
  }
intros k Hfst.
apply IH2.
apply wf_cn; auto.
eapply ofs_fst; eauto.
}

(* sigma *)
{
intros T1 IH1 T2 IH2 G Hwf.
rephrase aofs (aofs G T1 /\ exists k, fsts T1 k /\ aofs (G; cl_cn k) T2).
apply dec_and_dep; [apply IH1; auto |].
intros HaofT1.
so (aofs_sound _ _ Hwf HaofT1) as HofT1.
apply dec_chain; eauto using fsts_fun.
  {
  so (can_fsts _ _ HofT1) as (k & Hfst).
  left; eauto.
  }
intros k Hfst.
apply IH2.
apply wf_cn; auto.
eapply ofs_fst; eauto.
}

(* circ *)
{
intros P IH G Hwf.
rephrase aofs (aofps G P).
apply IH; auto.
}

(* down *)
{
intros T IH G Hwf.
rephrase aofps (aofs G T).
apply IH; auto.
}

(* ex *)
{
intros k _ P IH G Hwf.
rephrase aofps (aofk G k /\ aofps (G; cl_cn k) P).
apply dec_and_dep; auto using aofk_decide.
intro Haofk.
so (aofk_sound _ _ Hwf Haofk) as Hofk.
apply IH.
apply wf_cn; auto.
}
Qed.


Lemma aofs_decide :
  forall G T,
    wf G
    -> decidable (aofs G T).
Proof.
intros G T Hwf.
exact (aofx_decide _ _ Hwf andel).
Qed.


Lemma aofps_decide :
  forall G P,
    wf G
    -> decidable (aofps G P).
Proof.
intros G T Hwf.
exact (aofx_decide _ _ Hwf ander).
Qed.


Lemma ofs_decide :
  forall G T,
    wf G
    -> decidable (ofs G T).
Proof.
intros G T Hwf.
apply (dec_equiv (aofs G T)); eauto using aofs_sound, aofs_complete, aofs_decide.
Qed.


Lemma ofps_decide :
  forall G P,
    wf G
    -> decidable (ofps G P).
Proof.
intros G T Hwf.
apply (dec_equiv (aofps G T)); eauto using aofps_sound, aofps_complete, aofps_decide.
Qed.


Fixpoint sgsize (T : term) {struct T} : nat :=
match T with
| sg_one => 0
| sg_stat _ => 0
| sg_dyn _ => 0
| sg_pi T1 T2 => 1 + sgsize T1 + sgsize T2
| sg_sigma T1 T2 => 1 + sgsize T1 + sgsize T2
| sg_circ P => 1 + sgsize P
| ps_down T => 1 + sgsize T
| ps_ex k P => 1 + sgsize P
| _ => 0
end.


Lemma subst_sgsize :
  forall s P, hps P -> sgsize (subst s P) = sgsize P.
Proof.
exploit
  (hsig_ind
     (fun T => forall s, sgsize (subst s T) = sgsize T)
     (fun P => forall s, sgsize (subst s P) = sgsize P)) as Hind;
try (intros; simp_sub; simpl; f_equal; eauto; done).
intros.
apply (Hind ander); auto.
Qed.


Lemma asubx_decide :
  (forall G T T',
     wf G
     -> ofs G T
     -> ofs G T'
     -> decidable (asubs G T T'))
  /\
  (forall G T T',
     wf G
     -> ofps G T
     -> ofps G T'
     -> decidable (asubps G T T')).
Proof.
cut (forall y,
       (forall G T1 T2,
          y = sgsize T1 + sgsize T2
          -> wf G
          -> ofs G T1
          -> ofs G T2
          -> decidable (asubs G T1 T2))
       /\
       (forall G T1 T2,
          y = sgsize T1 + sgsize T2
          -> wf G
          -> ofps G T1
          -> ofps G T2
          -> decidable (asubps G T1 T2))).
  {
  intro Hind.
  split.
    {
    intros G T1 T2 Hwf Hof1 Hof2.
    apply (Hind (sgsize T1 + sgsize T2)); auto.
    }

    {
    intros G T1 T2 Hwf Hof1 Hof2.
    apply (Hind (sgsize T1 + sgsize T2)); auto.
    }
  }
intro y.
wfinduct y using lt_wf.
intros y IH.
split.
  {
  intros G T1 T2 -> Hwf Hof1 Hof2.
  revert T2 IH Hwf Hof2.
  case Hof1; clear codependent Hof1.
    (* one *)
    {
    intros G T2 IH Hwf Hof2.
    invert Hof2; try (intros; right; intro HH; invert HH; intros; subst; discriminate).
    intros <-.
    left.
    apply asubs_one.
    }

    (* stat *)
    {
    intros G k1 Hofk1 T2 IH Hwf Hof2.
    invert Hof2; try (intros; right; intro HH; invert HH; intros; subst; discriminate).
    intros k2 Hofk2.
    intros <-.
    rephrase asubs (asubk G k1 k2).
    apply asubk_decide; auto.
    }

    (* dyn *)
    {
    intros G t1 Hoft1 T2 IH Hwf Hof2.
    invert Hof2; try (intros; right; intro HH; invert HH; intros; subst; discriminate).
    intros t2 Hoft2 <-.
    rephrase asubs (aeq G G t1 t2 kd_type kd_type).
    apply aeq_decide; auto.
    }

    (* pi *)
    {
    intros G T1 U1 k1 HofT1 Hfst1 HofU1 TU2 IH Hwf HofTU2.
    invert HofTU2; try (intros; right; intro HH; invert HH; intros; subst; discriminate).
    intros T2 U2 k2 HofT2 Hfst2 HofU2 <-.
    simpl in IH.
    rephrase asubs (asubs G T2 T1 /\ exists k2a, fsts T2 k2a /\ asubs (G; cl_cn k2a) U1 U2).
    apply dec_and_dep.
      {
      exploit (IH (sgsize T2 + sgsize T1)) as H; [omega |].
      exact (H andel _ _ _ (eq_refl _) Hwf HofT2 HofT1).
      }
    intro Hsub1.
    apply dec_chain; eauto using fsts_fun.
      {
      left.
      exists k2.
      auto.
      }
    intros k2a Hfst2a.
    so (fsts_fun _#3 Hfst2 Hfst2a); subst k2a.
    exploit (IH (sgsize U1 + sgsize U2)) as H; [omega |].
    so (subs_fst _#5 Hwf (asubs_sound _#3 Hwf HofT2 HofT1 Hsub1) Hfst2 Hfst1) as Hsub1'.
    so (subst_ofs _#4 (subof_subsume_0 _#3 Hsub1') HofU1) as HofU1'.
    simp_sub in HofU1'.
    so (ofs_fst _#3 HofT2 Hfst2) as Hofk2.
    exact (H andel _ _ _ (eq_refl _) (wf_cn Hwf Hofk2) HofU1' HofU2).
    }

    (* sigma *)
    {
    intros G T1 U1 k1 HofT1 Hfst1 HofU1 TU2 IH Hwf HofTU2.
    invert HofTU2; try (intros; right; intro HH; invert HH; intros; subst; discriminate).
    intros T2 U2 k2 HofT2 Hfst2 HofU2 <-.
    simpl in IH.
    rephrase asubs (asubs G T1 T2 /\ exists k1a, fsts T1 k1a /\ asubs (G; cl_cn k1a) U1 U2).
    apply dec_and_dep.
      {
      exploit (IH (sgsize T1 + sgsize T2)) as H; [omega |].
      exact (H andel _ _ _ (eq_refl _) Hwf HofT1 HofT2).
      }
    intro Hsub1.
    apply dec_chain; eauto using fsts_fun.
      {
      left.
      exists k1.
      auto.
      }
    intros k1a Hfst1a.
    so (fsts_fun _#3 Hfst1 Hfst1a); subst k1a.
    exploit (IH (sgsize U1 + sgsize U2)) as H; [omega |].
    so (subs_fst _#5 Hwf (asubs_sound _#3 Hwf HofT1 HofT2 Hsub1) Hfst1 Hfst2) as Hsub1'.
    so (subst_ofs _#4 (subof_subsume_0 _#3 Hsub1') HofU2) as HofU2'.
    simp_sub in HofU2'.
    so (ofs_fst _#3 HofT1 Hfst1) as Hofk1.
    exact (H andel _ _ _ (eq_refl _) (wf_cn Hwf Hofk1) HofU1 HofU2').
    }

    (* circ *)
    {
    intros G P1 HofP1 T2 IH Hwf Hof2.
    invert Hof2; try (intros; right; intro HH; invert HH; intros; subst; discriminate).
    intros P2 HofP2 <-.
    simpl in IH.
    rephrase asubs (asubps G P1 P2).
    exploit (IH (sgsize P1 + sgsize P2)) as H; [omega |].
    exact (H ander _#3 (eq_refl _) Hwf HofP1 HofP2).
    }
  }

  {
  intros G P1 P2 -> Hwf Hof1 Hof2.
  revert P2 IH Hwf Hof2.
  case Hof1; clear codependent Hof1.
    (* down *)
    {
    intros G T1 HofT1 P2 IH Hwf Hof2.
    invert Hof2; try (intros; right; intro HH; invert HH; intros; subst; discriminate).
    intros T2 HofT2 <-.
    rephrase asubps (asubs G T1 T2).
    simpl in IH.
    exploit (IH (sgsize T1 + sgsize T2)) as H; [omega |].
    exact (H andel _#3 (eq_refl _) Hwf HofT1 HofT2).
    }

    (* ex *)
    {
    intros G k P1 Hofk HofP1 P2 IH Hwf HofP2.
    rephrase asubps (asubps (G; cl_cn k) P1 (shift1 P2)).
    simpl in IH.
    exploit (IH (sgsize P1 + sgsize P2)) as H; [omega |].
    so (H ander (G; cl_cn k) P1 (shift1 P2)) as H'.
    lapply H'.
      {
      intro H''.
      apply H''; auto using wf.
      eapply subst_ofps; eauto.
      apply weaken1.
      }

      {
      unfold shift1.
      rewrite -> subst_sgsize; auto.
      eapply ofps_hps; eauto.
      }
    }
  }
Qed.


Lemma asubs_decide :
  forall G T T',
    wf G
    -> ofs G T
    -> ofs G T'
    -> decidable (asubs G T T').
Proof.
exact (asubx_decide andel).
Qed.


Lemma asubps_decide :
  forall G T T',
    wf G
    -> ofps G T
    -> ofps G T'
    -> decidable (asubps G T T').
Proof.
exact (asubx_decide ander).
Qed.


Lemma subs_decide :
  forall G T T',
    wf G
    -> syn T
    -> ana T'
    -> ofs G T
    -> ofs G T'
    -> decidable (subs G T T').
Proof.
intros G T T' Hwf Hsyn Hana HofT HofT'.
apply (dec_equiv (asubs G T T')); auto using asubs_decide, asubs_sound, asubs_complete.
Qed.


Lemma subps_decide :
  forall G P P',
    wf G
    -> synp P
    -> anap P'
    -> ofps G P
    -> ofps G P'
    -> decidable (subps G P P').
Proof.
intros G T T' Hwf Hsyn Hana HofT HofT'.
apply (dec_equiv (asubps G T T')); auto using asubps_decide, asubps_sound, asubps_complete.
Qed.


Lemma can_headnorm :
  forall G c,
    wf G
    -> ofc G c kd_type
    -> exists c', headnorm G c c'.
Proof.
intros G c Hwf Hofc.
so (aeq_complete _#4 Hwf (eqc_refl Hofc)) as Haeq.
invert Haeq.
intros p _ H _ _.
eauto.
Qed.


Lemma svalue_decide :
  forall e, decidable (svalue e).
Proof.
intros e.
induct e;
try (intros; left; eauto using svalue; done);
try (intros; right; intro HH; invert HH; done).

(* pair *)
{
intros e1 IH1 e2 IH2.
rephrase svalue (svalue e1 /\ svalue e2).
apply dec_and; auto.
}

(* pack *)
{
intros c _ e IH k _ t _.
rephrase svalue (svalue e).
auto.
}
Qed.


Lemma can_linv :
  forall P,
    hps P
    -> exists G T, linv P G T.
Proof.
intros P H.
induct H; eauto using linv.
(* ex *)
{
intros k P _ _ (G & T & H).
eauto using linv.
}
Qed.


Lemma can_close :
  forall G P,
    Forall (fun A => exists k, A = cl_cn k) G
    -> exists P', close G P P'.
Proof.
intros G P H.
revert P.
induct H; eauto using close.
(* cons *)
intros x l (k & ->) _ IH P.
so (IH (ps_ex k P)) as (P' & Hclose).
eauto using close.
Qed.


Lemma aprint_fun :
  forall G e t t',
    wf G
    -> aprint G e t
    -> aprint G e t'
    -> t = t'.
Proof.
intros G e t t' Hwf Hprin Hprin'.
eapply print_fun; eauto using aprint_sound_pre.
Qed.


Lemma aprins_fun :
  forall G M T T',
    wf G
    -> aprins G M T
    -> aprins G M T'
    -> T = T'.
Proof.
intros G e t t' Hwf Hprin Hprin'.
eapply prins_fun; eauto using aprins_sound_pre.
Qed.


Lemma aprinps_fun :
  forall G M T T',
    wf G
    -> aprinps G M T
    -> aprinps G M T'
    -> T = T'.
Proof.
intros G e t t' Hwf Hprin Hprin'.
eapply prinps_fun; eauto using aprinps_sound_pre.
Qed.


Lemma aprinx_decide :
  (forall G e,
     wf G
     -> user e
     -> decidable (exists t, aprint G e t))
  /\
  (forall G M,
     wf G
     -> user M
     -> decidable (exists T, aprins G M T))
  /\
  (forall G M,
     wf G
     -> user M
     -> decidable (exists T, aprinps G M T)).
Proof.
cut (forall m,
       user m
       -> (forall G s, wf G -> allvar s -> decidable (exists t, aprint G (subst s m) t))
          /\
          (forall G s, wf G -> allvar s -> decidable (exists T, aprins G (subst s m) T))
          /\
          (forall G s, wf G -> allvar s -> decidable (exists P, aprinps G (subst s m) P))).
  {
  intro Hind.
  do2 2 split.
    {
    intros G m Hwf Huser.
    so (Hind _ Huser andel _ id Hwf allvar_id) as H.
    simp_sub in H.
    exact H.
    }

    {
    intros G m Hwf Huser.
    so (Hind _ Huser anderl _ id Hwf allvar_id) as H.
    simp_sub in H.
    exact H.
    }

    {
    intros G m Hwf Huser.
    so (Hind _ Huser anderr _ id Hwf allvar_id) as H.
    simp_sub in H.
    exact H.
    }
  }
intros m Huser.
induct Huser;
intros;
do2 2 split;
try (simp_sub; right; intro HH; destruct HH as (? & HH); simp_sub in HH; invert HH; done);
try (left; eexists; eauto using aprint, aprins; done);
revert_all.

(* var *)
{
intros i G s Hwf Hs.
simp_sub.
so (Hs i) as (i' & Heq).
rewrite -> Heq.
rephrase aprint (exists A, index i' G A /\ exists t, A = cl_tm t).
apply dec_chain; eauto using index_fun, index_decide.
intros A ?.
decform A.
}

(* var *)
{
intros i G s Hwf Hs.
simp_sub.
so (Hs i) as (i' & Heq).
rewrite -> Heq.
rephrase aprins (exists A, index i' G A /\ exists T c, A = cl_md T c).
apply dec_chain; eauto using index_fun, index_decide.
intros A ?.
decform A.
}

(* var *)
{
intros i G s _ Hs.
simp_sub.
so (Hs i) as (i' & Heq).
rewrite -> Heq.
right.
intros (v & H).
invert H.
}

(* lam *)
{
intros t e _ (IH, _) G s Hwf Hs.
simp_sub.
rephrase aprint (aofc G (subst s t) kd_type /\ exists t', aprint (G; cl_tm (subst s t)) (subst (dot (var 0) (compose s sh1)) e) t').
apply dec_and_dep; auto using aofc_decide with static.
intro Hoft.
apply IH; auto using wf, allvar_bind, aofc_sound with static.
}

(* app *)
{
intros e1 e2 _ (IH1, _) _ (IH2, _) G s Hwf Hs.
simp_sub.
apply (dec_equiv (exists t, aprint G (subst s e1) t /\ exists t', headnorm G t t' /\ exists t1 t2, t' = cn_arrow t1 t2 /\ exists t1', aprint G (subst s e2) t1' /\ aeqc G t1' t1 kd_type)).
  {
  into aprint.
  }
  
  {
  intro HH; destruct_all HH; invert HH; intros; eauto 10.
  }
apply dec_chain; eauto using aprint_fun.
intros t Hprin1.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin1)) as Hoft.
apply dec_chain; eauto using headnorm_fun.
  { left; apply can_headnorm; auto. }
intros t' Hnorm.
apply dec_chain2; [injfun | decform t' |].
intros t1 t2 ->.
apply dec_chain; eauto using aprint_fun.
intros t1' Hprin2.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin2)) as Hoft1'.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hoft Hnorm) anderl)) as (l & H).
invert H.
intros Hof1 _ _.
apply aeqc_decide; auto.
}

(* pair *)
{
intros e1 e2 _ (IH1, _) _ (IH2, _) G s Hwf Hs.
simp_sub.
rephrase aprint ((exists t1, aprint G (subst s e1) t1) /\ exists t2, aprint G (subst s e2) t2).
apply dec_and; auto.
}

(* pi1 *)
{
intros e _ (IH, _) G s Hwf Hs.
simp_sub.
rephrase aprint (exists t, aprint G (subst s e) t /\ exists t', headnorm G t t' /\ exists t1 t2, t' = cn_prod t1 t2).
apply dec_chain; eauto using aprint_fun.
intros t Hprin.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin)) as Hoft.
apply dec_chain; eauto using headnorm_fun.
  { left; apply can_headnorm; auto. }
intros t' _.
decform t'.
}

(* pi2 *)
{
intros e _ (IH, _) G s Hwf Hs.
simp_sub.
rephrase aprint (exists t, aprint G (subst s e) t /\ exists t', headnorm G t t' /\ exists t1 t2, t' = cn_prod t1 t2).
apply dec_chain; eauto using aprint_fun.
intros t Hprin.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin)) as Hoft.
apply dec_chain; eauto using headnorm_fun.
  { left; apply can_headnorm; auto. }
intros t' _.
decform t'.
}

(* plam *)
{
intros k e _ (IH, _) G s Hwf Hs.
simp_sub.
rephrase aprint (aofk G (subst s k) /\ svalue (subst (dot (var 0) (compose s sh1)) e) /\ exists t', aprint (G; cl_cn (subst s k)) (subst (dot (var 0) (compose s sh1)) e) t').
apply dec_and_dep; auto using aofk_decide.
intro Hofk.
apply dec_and; auto using svalue_decide.
apply IH; auto using wf, synctx, allvar_bind, aofk_sound.
}

(* papp *)
{
intros e c _ (IH, _) G s Hwf Hs.
simp_sub.
rephrase aprint (exists u, aprint G (subst s e) u /\ exists u', headnorm G u u' /\ exists k t, u' = cn_all k t /\ aofc G (subst s c) k).
apply dec_chain; eauto using aprint_fun.
intros u Hprin.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin)) as Hofu.
apply dec_chain; eauto using headnorm_fun.
  { left; apply can_headnorm; auto. }
intros u' Hnorm.
apply dec_chain2; [injfun | decform u' |].
intros k t ->.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hofu Hnorm) anderl)) as (l & H).
invert H.
intros Hofk _ _.
apply aofc_decide; auto.
}

(* pack *)
{
intros k t c e _ (IH, _) G s Hwf Hs.
simp_sub.
apply (dec_equiv (aofk G (subst s k) /\ aofc G (subst s c) (subst s k) /\ exists u, aprint G (subst s e) u /\ aofc (G; cl_cn (subst s k)) (subst (dot (var 0) (compose s sh1)) t) kd_type /\ aeqc G u (subst1 (subst s c) (subst (dot (var 0) (compose s sh1)) t)) kd_type)).
  { into aprint. }

  {
  intros (v & Hprin).
  invert Hprin.
  intros u Hofk Hofc Hprine Heqc Hoft <-.
  eauto 8.
  }
apply dec_and_dep; auto using aofk_decide.
intro Haofk.
so (aofk_sound _ _ Hwf Haofk) as Hofk.
apply dec_and_dep; auto using aofc_decide.
intro Hofc.
apply dec_chain; eauto using aprint_fun.
intros u Hprin.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin)) as Hofu.
apply dec_and_dep.
  {
  apply aofc_decide; auto with static.
  }
intros Hoft.
apply aeqc_decide; auto.
eapply subst1_ofc_type; eauto using aofc_sound, wf with static.
}

(* unpack *)
{
intros e1 e2 _ (IH1, _) _ (IH2, _) G s Hwf Hs.
simp_sub.
apply (dec_equiv (exists u, aprint G (subst s e1) u /\ exists u', headnorm G u u' /\ exists k t, u' = cn_exists k t /\ exists t', aprint (G; cl_cn k; cl_tm t) (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) e2) t' /\ aeqc (G; cl_cn k) (subst (dot nonsense id) t') (subst (dot nonsense (dot (shift1 (inh k)) sh1)) t') kd_type)).
  { into aprint. }

  {
  intros (v & Hprin).
  invert Hprin.
  intros; subst.
  eauto 10.
  }
apply dec_chain; eauto using aprint_fun.
intros u Hprin1.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin1)) as Hofu.
apply dec_chain; eauto using headnorm_fun.
  { left; apply can_headnorm; auto. }
intros u' Hnorm.
apply dec_chain2; [injfun | decform u' |].
intros k t ->.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hofu Hnorm) anderl)) as (l & H).
invert H.
intros Hofk Hoft _.
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
apply dec_chain; eauto using aprint_fun.
  {
  apply IH2; auto using synctx, allvar_bind2.
  }
intros t' Hprin2.
so (oft_reg _#3 Hwf' (aprint_sound _#3 Hwf' Hprin2)) as Hoft'.
apply aeqc_decide; auto using wf.
  {
  eapply subst_ofc_type; eauto.
  apply subof_dot; auto using subof_id.
  simp_sub.
  apply sof_tm.
  }

  {
  eapply subst_ofc_type; eauto.
  apply subof_dot.
    {
    apply subof_dot; eauto using subof_sh with sequence.
    simp_sub.
    apply sof_cn.
    apply weaken1_ofc.
    apply inh_ofc; auto.
    }

    {
    simp_sub.
    apply sof_tm.
    }
  }
}

(* fix *)
{
intros t e _ (IH, _) G s Hwf Hs.
simp_sub.
apply (dec_equiv (aofc G (subst s t) kd_type /\ exists u, aprint G (subst s e) u /\ aeqc G u (cn_arrow (cn_arrow cn_unit (subst s t)) (subst s t)) kd_type)).
  { into aprint. }

  {
  intros (v & Hprin).
  invert Hprin.
  intros u Hofc Hprin1 Heqc <-.
  so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin)) as Hoft.
  eauto.
  }
apply dec_and_dep; auto using aofc_decide with static.
intros Hoft.
apply dec_chain; eauto using aprint_fun.
intros u Hprin.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin)) as Hofu.
apply aeqc_decide; eauto using aofc_sound with static.
}

(* lett *)
{
intros e1 e2 _ (IH1, _) _ (IH2, _) G s Hwf Hs.
simp_sub.
rephrase aprint (exists t1, aprint G (subst s e1) t1 /\ exists t2, aprint (G; cl_tm t1) (subst (dot (var 0) (compose s sh1)) e2) t2).
apply dec_chain; eauto using aprint_fun.
intros t1 Hprin1.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin1)) as Hoft1.
eauto using wf, synctx, allvar_bind.
}

(* letm *)
{
intros M e HM (_ & _ & IH1) _ (IH2, _) G s Hwf Hs.
simp_sub.
simpl plus.
apply (dec_equiv (exists P, aprinps G (subst s M) P /\ exists G' T, linv P G' T /\ exists k, fsts T k /\ exists t, aprint (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) (subst (under 2 (sh (length G'))) (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) e)) t /\ aeqc (G;; G'; cl_cn k) (subst (dot nonsense id) t) (subst (dot nonsense (dot (subst (compose (cinh G') (sh (S (length G')))) (inh k)) (compose (cinh G') (sh (S (length G')))))) t) kd_type)).
  { into aprint. }

  {
  intros (v & H); invert H; intros; subst; eauto 15.
  }
apply dec_chain; eauto using aprinps_fun.
intros P Hprin1.
so (oflm_reg _#3 Hwf (aprinps_sound _#3 Hwf Hprin1)) as HofP.
so (ofps_hps _ _ HofP) as HP.
apply dec_chain2; eauto using linv_fun.
  {
  left.
  apply can_linv; auto.
  }
intros G' T Hinv.
so (linv_reg_gen _#4 Hwf HofP Hinv) as (Hwfi' & HofT).
so (wf_wfh_append _ _ Hwf Hwfi') as Hwf'.
apply dec_chain; eauto using fsts_fun.
  {
  so (can_fsts _ _ HofT) as (k & Hfst).
  left; eauto.
  }
intros k Hfst.
so (wf_md Hwf' HofT Hfst) as Hwf''.
so (ofs_fst _#3 HofT Hfst) as Hofk.
apply dec_chain; eauto using aprint_fun.
  {
  simp_sub.
  apply IH2; auto using wf, allvar_dot, allvar_sh, allvar_compose.
  }
intros t Hprin2.
so (oft_reg _#3 Hwf'' (aprint_sound _#3 Hwf'' Hprin2)) as Hoft.
apply aeqc_decide; auto using wf.
  {
  eapply subst_ofc_type; eauto.
  apply subof_dot; auto using subof_id.
  simp_sub.
  apply sof_md.
  }

  {
  eapply subst_ofc_type; eauto.
  assert (subof sof (G;; G'; cl_cn k) (compose (cinh G') (sh (S (length G')))) (G;; G')) as Hsinh.
    {
    apply (subof_sof_trans _ G); eauto using cinh_subof.
    rewrite -> app_comm_cons.
    replace (S (length G')) with (length (G'; cl_cn k)) by reflexivity.
    apply subof_sh.
    apply app_truncate.
    }
  apply subof_dot.
    {
    apply subof_dot; auto.
    simp_sub.
    apply sof_cn.
    eapply subst_ofc; eauto.
    apply inh_ofc; auto.
    }

    {
    simp_sub.
    apply sof_md.
    }
  }
}

(* ext *)
{
intros M _ (_ & IH & _) G s Hwf Hs.
simp_sub.
rephrase aprint (exists T, aprins G (subst s M) T /\ exists t, T = sg_dyn t).
apply dec_chain; eauto using aprins_fun.
intros T _.
decform T.
}

(* stat *)
{
intros c G s Hwf Hs.
simp_sub.
rephrase aprins (exists k, aprink G (subst s c) k).
apply aprink_decide; auto.
}

(* dyn *)
{
intros e _ (IH, _) G s Hwf Hs.
simp_sub.
rephrase aprins (exists t, aprint G (subst s e) t).
auto.
}

(* lam *)
{
intros T1 M HT _ (_ & IH & _) G s Hwf Hs.
simp_sub.
rephrase aprins (aofs G (subst s T1) /\ exists k1, fsts (subst s T1) k1 /\ exists T2, aprins (G; cl_cn k1; cl_md (shift1 (subst s T1)) (var 0)) (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) M) T2 /\ exists T2', T2 = shift1 T2').
apply dec_and_dep.
  {
  apply aofs_decide; auto.
  }
intro HaofT1.
apply dec_chain; eauto using fsts_fun.
  {
  so (can_fsts _ _ (aofs_sound _ _ Hwf HaofT1)) as (k1 & Hfst).
  left; eauto.
  }
intros k1 Hfst.
so (aofs_sound _ _ Hwf HaofT1) as HofT1.
apply dec_chain; eauto using aprins_fun, wf.
  {
  apply IH; auto using wf, allvar_bind2.
  }
intros T2 Hprin.
so (wf_md Hwf HofT1 Hfst) as Hwf'.
so (ofm_reg _#3 Hwf' (aprins_sound _#3 Hwf' Hprin)) as HofT2.
left.
exists (subst1 nonsense T2).
simp_sub.
eapply subord_md_sg; eauto.
}

(* app *)
{
intros M1 M2 Huser1 (_ & IH1 & _) Huser2 (_ & IH2 & _) G s Hwf Hs.
simp_sub.
apply (dec_equiv (exists T, aprins G (subst s M1) T /\ exists T1 T2, T = sg_pi T1 T2 /\ exists T1', aprins G (subst s M2) T1' /\ asubs G T1' T1 /\ exists c2, fstm G (subst s M2) c2)).
  { into aprins. }

  { 
  intro H; destruct_all H; invert H; intros; subst; eauto 10.
  }
apply dec_chain; eauto using aprins_fun.
intros T Hprin1.
apply dec_chain2; [injfun | decform T |].
intros T1 T2 ->.
apply dec_chain; eauto using aprins_fun.
intros T1' Hprin2.
so (ofm_reg _#3 Hwf (aprins_sound _#3 Hwf Hprin1)) as Hofpi.
invert Hofpi.
intros _ HofT1 _ _.
so (ofm_reg _#3 Hwf (aprins_sound _#3 Hwf Hprin2)) as HofT1'.
apply dec_and.
  {
  apply asubs_decide; auto.
  }
so (can_fstm _#3 Hwf (aprins_sound _#3 Hwf Hprin2)).
left; auto.
}

(* pair *)
{
intros M1 M2 _ (_ & IH1 & _) _ (_ & IH2 & _) G s Hwf Hs.
simp_sub.
rephrase aprins ((exists T1, aprins G (subst s M1) T1) /\ exists T2, aprins G (subst s M2) T2).
apply dec_and; auto.
}

(* pi1 *)
{
intros M _ (_ & IH & _) G s Hwf Hs.
simp_sub.
rephrase aprins (exists T, aprins G (subst s M) T /\ exists T1 T2, T = sg_sigma T1 T2).
apply dec_chain; eauto using aprins_fun.
intros T _.
decform T.
}

(* pi2 *)
{
intros M _ (_ & IH & _) G s Hwf Hs.
simp_sub.
rephrase aprins (exists T, aprins G (subst s M) T /\ (exists T1 T2, T = sg_sigma T1 T2) /\ exists c, fstm G (subst s M) c).
apply dec_chain; eauto using aprins_fun.
intros T Hprin.
apply dec_and.
  { decform T. }
so (can_fstm _#3 Hwf (aprins_sound _#3 Hwf Hprin)).
left; auto.
}

(* lett *)
{
intros e M _ (IH1 & _) _ (_ & IH2 & _) G s Hwf Hs.
simp_sub.
rephrase aprins (exists t, aprint G (subst s e) t /\ exists T, aprins (G; cl_tm t) (subst (dot (var 0) (compose s sh1)) M) T).
apply dec_chain; eauto using aprint_fun.
intros t Hprin.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin)) as Hoft.
apply IH2; auto using wf, allvar_bind.
}

(* cir *)
{
intros M _ (_ & _ & IH) G s Hwf Hs.
simp_sub.
rephrase aprins (exists P, aprinps G (subst s M) P).
auto.
}

(* ret *)
{
intros M _ (_ & IH & _) G s Hwf Hs.
simp_sub.
rephrase aprinps (exists T, aprins G (subst s M) T).
auto.
}

(* seal *)
{
intros M T Huser (_ & IH & _) HT G s Hwf Hs.
simp_sub.
rephrase aprinps (aofs G (subst s T) /\ exists T', aprins G (subst s M) T' /\ asubs G T' (subst s T)).
apply dec_and_dep; auto using aofs_decide.
intro HofT.
apply dec_chain; eauto using aprins_fun.
intros T' Hprin.
so (ofm_reg _#3 Hwf (aprins_sound _#3 Hwf Hprin)) as HofT'.
apply asubs_decide; auto using aofs_sound.
}

(* bind *)
{
intros M1 M2 Huser1 (_ & IH1 & _) _ (_ & _ & IH2) G s Hwf Hs.
simp_sub.
apply (dec_equiv (exists P, aprins G (subst s M1) P /\ exists P1, P = sg_circ P1 /\ exists G' T, linv P1 G' T /\ exists k, fsts T k /\ exists P2, aprinps (G;; G'; cl_cn k; cl_md (shift1 T) (var 0)) (subst (under 2 (sh (length G'))) (subst (dot (var 0) (dot (var 1) (compose s (sh 2))))  M2)) P2 /\ exists P2', close G' (ps_ex k (subst1 nonsense P2)) P2')).
  { into aprinps. }

  {
  intros (v & H); invert H; intros; subst; eauto 15.
  }
apply dec_chain; eauto using aprins_fun.
intros P Hprin1.
apply dec_chain; [injfun | decform P |].
intros P1 ->.
so (ofm_reg _#3 Hwf (aprins_sound _#3 Hwf Hprin1)) as Hofcirc.
invert Hofcirc.
intro HofP1.
so (ofps_hps _ _ HofP1) as Hps.
apply dec_chain2; eauto using linv_fun.
  {
  left.
  apply can_linv; auto.
  }
intros G' T Hinv.
so (linv_reg _#4 Hwf HofP1 Hinv) as (Hwf' & HofT).
apply dec_chain; eauto using fsts_fun.
  {
  so (can_fsts _ _ HofT) as (k & Hfst).
  left; eauto.
  }
intros k Hfst.
apply dec_chain; eauto using aprinps_fun, wf.
  {
  simp_sub.
  apply IH2; auto using wf, allvar_dot, allvar_sh, allvar_compose.
  }
intros P2 Hprin2.
left.
apply can_close.
eapply linv_allcn; eauto.
}

(* unpack *)
{
intros e M _ (IH1 & _) _ (_ & _ & IH2) G s Hwf Hs.
simp_sub.
apply (dec_equiv (exists u, aprint G (subst s e) u /\ exists u', headnorm G u u' /\ exists k t, u' = cn_exists k t /\ exists P, aprinps (G; cl_cn k; cl_tm t) (subst (dot (var 0) (dot (var 1) (compose s (sh 2)))) M) P)).
  { into aprinps. }

  {
  intros (v & H); invert H; intros; eauto 10.
  }
apply dec_chain; eauto using aprint_fun.
intros u Hprin1.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin1)) as Hofu.
apply dec_chain; eauto using headnorm_fun.
  { left; apply can_headnorm; auto. }
intros u' Hnorm.
apply dec_chain2; [injfun | decform u' |].
intros k t ->.
so (ofc_ofcr _#3 (eqc_reg _#4 Hwf (headnorm_sound _#4 Hwf Hofu Hnorm) anderl)) as (l & H).
invert H.
intros Hofk Hoft _.
so (wf_tm (wf_cn Hwf Hofk) Hoft) as Hwf'.
apply IH2; auto using allvar_bind2, synctx.
}
Qed.


Theorem aprint_decide :
  forall G e,
     wf G
     -> user e
     -> decidable (exists t, aprint G e t).
Proof.
exact (aprinx_decide andel).
Qed.


Theorem aprins_decide :
  forall G M,
     wf G
     -> user M
     -> decidable (exists T, aprins G M T).
Proof.
exact (aprinx_decide anderl).
Qed.


Theorem aprinps_decide :
  forall G M,
     wf G
     -> user M
     -> decidable (exists P, aprinps G M P).
Proof.
exact (aprinx_decide anderr).
Qed.


Corollary oft_some_decide :
  forall G e,
    wf G
    -> synctx G
    -> user e
    -> decidable (exists t, oft G e t).
Proof.
intros G e Hwf HG Huser.
apply (dec_equiv (exists t, aprint G e t)); auto using aprint_decide.
  {
  intros (t & Hprin).
  exists t.
  apply aprint_sound; auto.
  }

  {
  intros (t & Hof).
  so (aprint_complete _#3 Hwf HG Huser Hof) as (t' & Hprin & _).
  eauto.
  }
Qed.


Corollary ofm_some_decide :
  forall G M,
    wf G
    -> synctx G
    -> user M
    -> decidable (exists T, ofm G M T).
Proof.
intros G M Hwf HG Huser.
apply (dec_equiv (exists T, aprins G M T)); auto using aprins_decide.
  {
  intros (T & Hprin).
  exists T.
  apply aprins_sound; auto.
  }

  {
  intros (T & Hof).
  so (aprins_complete _#3 Hwf HG Huser Hof) as (t' & Hprin & _).
  eauto.
  }
Qed.


Corollary oflm_some_decide :
  forall G M,
    wf G
    -> synctx G
    -> user M
    -> decidable (exists P, oflm G M P).
Proof.
intros G M Hwf HG Huser.
apply (dec_equiv (exists T, aprinps G M T)); auto using aprinps_decide.
  {
  intros (T & Hprin).
  exists T.
  apply aprinps_sound; auto.
  }

  {
  intros (T & Hof).
  so (aprinps_complete _#3 Hwf HG Huser Hof) as (t' & Hprin & _).
  eauto.
  }
Qed.


Corollary oft_decide :
  forall G e t,
    wf G
    -> synctx G
    -> user e
    -> decidable (oft G e t).
Proof.
intros G e t Hwf HG He.
apply (dec_equiv (ofc G t kd_type /\ exists t', aprint G e t' /\ eqc G t' t kd_type)).
  {
  intros (Hoft & t' & Hprin & Heqt).
  so (aprint_sound _#3 Hwf Hprin) as Hofe.
  eauto with static.
  }

  {
  intros Hofe.
  split.
    {
    eapply oft_reg; eauto.
    }
  so (aprint_complete _#3 Hwf HG He Hofe) as (t' & Hprin & Heqt).
  eauto.
  }
apply dec_and_dep; eauto using ofc_decide with static.
intro Hoft.
apply dec_chain; eauto using aprint_fun, aprint_decide.
intros t' Hprin.
so (oft_reg _#3 Hwf (aprint_sound _#3 Hwf Hprin)) as Hoft'.
apply eqc_decide; auto.
Qed.


Corollary ofm_decide :
  forall G M T,
    wf G
    -> synctx G
    -> user M
    -> ana T
    -> decidable (ofm G M T).
Proof.
intros G M T Hwf HG HM HT.
apply (dec_equiv (ofs G T /\ exists T', aprins G M T' /\ subs G T' T)).
  {
  intros (HofT & T' & Hprin & Hsub).
  so (aprins_sound _#3 Hwf Hprin) as HofM.
  eauto with static.
  }

  {
  intros HofM.
  split.
    {
    eapply ofm_reg; eauto.
    }
  so (aprins_complete _#3 Hwf HG HM HofM) as (T' & Hprin & Hsub).
  eauto.
  }
apply dec_and_dep; eauto using ofs_decide with static.
intro HofT.
apply dec_chain; eauto using aprins_fun, aprins_decide.
intros T' Hprin.
so (ofm_reg _#3 Hwf (aprins_sound _#3 Hwf Hprin)) as HofT'.
apply subs_decide; auto.
eapply aprins_syn; eauto.
Qed.


Corollary oflm_decide :
  forall G M P,
    wf G
    -> synctx G
    -> user M
    -> anap P
    -> decidable (oflm G M P).
Proof.
intros G M T Hwf HG HM HT.
apply (dec_equiv (ofps G T /\ exists T', aprinps G M T' /\ subps G T' T)).
  {
  intros (HofT & T' & Hprin & Hsub).
  so (aprinps_sound _#3 Hwf Hprin) as HofM.
  eauto with static.
  }

  {
  intros HofM.
  split.
    {
    eapply oflm_reg; eauto.
    }
  so (aprinps_complete _#3 Hwf HG HM HofM) as (T' & Hprin & Hsub).
  eauto.
  }
apply dec_and_dep; eauto using ofps_decide with static.
intro HofT.
apply dec_chain; eauto using aprinps_fun, aprinps_decide.
intros T' Hprin.
so (oflm_reg _#3 Hwf (aprinps_sound _#3 Hwf Hprin)) as HofT'.
apply subps_decide; auto.
eapply aprinps_syn; eauto.
Qed.
