
Require Import Tactics.

Require Import Subst.
Require Import Static.
Require Import Formation.



Inductive skind : Set :=
| sk_type  : skind
| sk_pi    : skind -> skind -> skind
| sk_sigma : skind -> skind -> skind
| sk_one   : skind
.


Fixpoint simp (k : (* kd *) term) {struct k} : skind :=
  (match k with
   | kd_type => sk_type
   | kd_sing c => sk_type
   | kd_pi k1 k2 => sk_pi (simp k1) (simp k2)
   | kd_sigma k1 k2 => sk_sigma (simp k1) (simp k2)
   | _ => sk_one
   end).


Lemma subst_simp :
  forall s k, hkd k -> simp (subst s k) = simp k.
Proof.
intros s k Hkd.
revert s.
induct Hkd; intros; simp_sub; simpl; try f_equal; auto.
Qed.


Lemma eqk_simp :
  forall G k k',
    eqk G k k' -> simp k = simp k'.
Proof.
intros G k k' Heqk.
induct Heqk.

(* refl *)
intros; reflexivity.

(* symm *)
intros; symmetry; assumption.

(* trans *)
intros; etransitivity; eassumption.

(* sing *)
intros; simpl; reflexivity.

(* pi *)
intros; simpl; f_equal; auto; done.

(* sigma *)
intros; simpl; f_equal; auto; done.
Qed.


Lemma subk_simp :
  forall G k k',
    subk G k k' -> simp k = simp k'.
Proof.
intros G k k' Hsubk.
induct Hsubk.

(* refl *)
intros.
eapply eqk_simp; eauto; done.

(* trans *)
intros; etransitivity; eassumption.

(* sing_t *)
intros; simpl; reflexivity.

(* pi *)
intros; simpl; f_equal; auto; done.

(* sigma *)
intros; simpl; f_equal; auto; done.
Qed.
